# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
from PIL import Image, ImageChops
from pelican.tests.support import unittest, get_settings, temporary_folder


from image_process import harvest_images_in_fragment, process_image

try:
    import unittest.mock as mock  # python < 3.3
except ImportError:
    import mock


CUR_DIR = os.path.dirname(__file__)
TEST_IMAGES = [os.path.join(CUR_DIR, 'test_data/pelican-bird.jpg'),
               os.path.join(CUR_DIR, 'test_data/pelican-bird.png')]


class ImageDerivativeTest(unittest.TestCase):
    transforms = {
        'crop': ['crop 10 20 100 200'],
        'flip_horizontal': ['flip_horizontal'],
        'flip_vertical': ['flip_vertical'],
        'grayscale': ['grayscale'],
        'resize': ['resize 200 250'],
        'rotate': ['rotate 20'],
        'scale_in': ['scale_in 200 250 False'],
        'scale_out': ['scale_out 200 250 False'],
        'blur': ['blur'],
        'contour': ['contour'],
        'detail': ['detail'],
        'edge_enhance': ['edge_enhance'],
        'edge_enhance_more': ['edge_enhance_more'],
        'emboss': ['emboss'],
        'find_edges': ['find_edges'],
        'smooth': ['smooth'],
        'smooth_more': ['smooth_more'],
        'sharpen': ['sharpen'],
        }

    @mock.patch('image_process.process_image')
    def test_extraction(self, process_image):

        settings = get_settings(
            IMAGE_PROCESS_DIR='derivatives', IMAGE_PROCESS=self.transforms
        )
        html = ('<img class="test image-process image-process-crop test2"'
                ' src="/tmp/test.jpg" />')

        html = harvest_images_in_fragment(html, settings)

        expected_content = ('<img class="test image-process image-process-crop'
                            ' test2" src="/tmp/derivatives/crop/test.jpg"/>')

        expected_source = os.path.join(settings['PATH'], 'tmp/test.jpg')
        expected_destination = os.path.join(
            settings['OUTPUT_PATH'], 'tmp', settings['IMAGE_PROCESS_DIR'],
            'crop', 'test.jpg'
        )
        expected_image = (expected_source, expected_destination,
                          ['crop 10 20 100 200'])
        expected_calls = [mock.call(expected_image, settings)]

        self.assertEqual(html, expected_content)
        self.assertEqual(expected_calls, process_image.call_args_list)

    def test_transforms(self):
        settings = get_settings(IMAGE_PROCESS=self.transforms)

        def test_transform(d, i, tmpdir):
            path, name = os.path.split(i)
            destination = os.path.join(tmpdir, d, name)
            image = (i, destination, settings['IMAGE_PROCESS'][d])

            process_image(image, settings)

            transformed = Image.open(destination)

            expected_path = os.path.join(path, 'results', d, name)
            expected = Image.open(expected_path)

            img_diff = ImageChops.difference(transformed, expected).getbbox()
            self.assertEqual(img_diff, None)

        with temporary_folder() as tmpdir:
            [test_transform(d, i, tmpdir)
             for d in self.transforms for i in TEST_IMAGES]


class HTMLGenerationTest(unittest.TestCase):
    """
    Check that all syntaxes generate proper HTML.
    """
    valid_transforms = {
        'thumb': ["crop 0 0 50% 50%", "scale_out 150 150", "crop 0 0 150 150"],
        'article-image': {
            'type': 'image', 'ops': ["scale_in 300 300"],
        },
        'crisp': {
            'type': 'responsive-image',
            'srcset': [('1x', ["scale_in 800 600 True"]),
                       ('2x', ["scale_in 1600 1200 True"]),
                       ('4x', ["scale_in 3200 2400 True"])],
            'default': '1x',
        },
        'crisp2': {
            'type': 'responsive-image',
            'srcset': [('1x', ["scale_in 800 600 True"]),
                       ('2x', ["scale_in 1600 1200 True"]),
                       ('4x', ["scale_in 3200 2400 True"])],
            'default': ["scale_in 400 300 True"],
        },
        'large-photo': {
            'type': 'responsive-image',
            'sizes': (
                '(min-width: 1200px) 800px, (min-width: 992px) 650px, '
                '(min-width: 768px) 718px, 100vw'
            ),
            'srcset': [('600w', ["scale_in 600 450 True"]),
                       ('800w', ["scale_in 800 600 True"]),
                       ('1600w', ["scale_in 1600 1200 True"])],
            'default': '800w',
        },
        'pict': {
            'type': 'picture',
            'sources': [
                {
                    'name': 'default',
                    'media': '(min-width: 640px)',
                    'srcset': [('640w', ["scale_in 640 480 True"]),
                               ('1024w', ["scale_in 1024 683 True"]),
                               ('1600w', ["scale_in 1600 1200 True"])],
                    'sizes': '100vw',
                }, {
                    'name': 'source-1',
                    'srcset': [('1x', ["crop 100 100 200 200"]),
                               ('2x', ["crop 100 100 300 300"])]
                }
            ],
            'default': ('default', '640w'),
        },
        'pict2': {
            'type': 'picture',
            'sources': [
                {
                    'name': 'default',
                    'media': '(min-width: 640px)',
                    'srcset': [('640w', ["scale_in 640 480 True"]),
                               ('1024w', ["scale_in 1024 683 True"]),
                               ('1600w', ["scale_in 1600 1200 True"])],
                    'sizes': '100vw'
                }, {
                    'name': 'source-2',
                    'srcset': [('1x', ["crop 100 100 200 200"]),
                               ('2x', ["crop 100 100 300 300"])]
                }
            ],
            'default': ('source-2', ["scale_in 800 600 True"]),
        },
    }

    def test_undefined(self):
        settings = get_settings(IMAGE_PROCESS=self.valid_transforms)
        html = '<img class="image-process-undefined" src="/tmp/test.jpg" />'
        with self.assertRaises(RuntimeError):
            harvest_images_in_fragment(html, settings)

    @mock.patch('image_process.process_image')
    def test_image_generation(self, process_image):
        settings = get_settings(IMAGE_PROCESS=self.valid_transforms,
                                IMAGE_PROCESS_DIR='derivs')

        test_data = [
            ('<img class="image-process-thumb" src="/tmp/test.jpg" />',
             '<img class="image-process-thumb" '
             'src="/tmp/derivs/thumb/test.jpg"/>',
             (
                 'tmp/test.jpg', 'thumb/test.jpg',
                 ["crop 0 0 50% 50%", "scale_out 150 150", "crop 0 0 150 150"]
             )),
            ('<img class="image-process-article-image" src="/tmp/test.jpg" />',
             '<img class="image-process-article-image" '
             'src="/tmp/derivs/article-image/test.jpg"/>',
             ('tmp/test.jpg', 'article-image/test.jpg', ["scale_in 300 300"]))
        ]

        for data in test_data:
            expected_source = os.path.join(settings['PATH'], data[2][0])
            expected_destination = os.path.join(
                settings['OUTPUT_PATH'], 'tmp', settings['IMAGE_PROCESS_DIR'],
                data[2][1]
            )
            html = harvest_images_in_fragment(data[0], settings)

            expected_image = (expected_source, expected_destination,
                              data[2][2])

            expected_calls = [mock.call(expected_image, settings)]
            self.assertEqual(html, data[1])
            self.assertEqual(expected_calls, process_image.call_args_list)
            process_image.reset_mock()

    @mock.patch('image_process.process_image')
    def test_responsive_image_generation(self, process_image):
        settings = get_settings(IMAGE_PROCESS=self.valid_transforms,
                                IMAGE_PROCESS_DIR='derivs')
        test_data = [
            ('<img class="image-process-crisp" src="/tmp/test.jpg" />',
             '<img class="image-process-crisp" '
             'src="/tmp/derivs/crisp/1x/test.jpg" '
             'srcset="/tmp/derivs/crisp/1x/test.jpg 1x, '
             '/tmp/derivs/crisp/2x/test.jpg 2x, '
             '/tmp/derivs/crisp/4x/test.jpg 4x"/>',

             [('tmp/test.jpg', 'crisp/1x/test.jpg',
               ["scale_in 800 600 True"]),
              ('tmp/test.jpg', 'crisp/2x/test.jpg',
               ["scale_in 1600 1200 True"]),
              ('tmp/test.jpg', 'crisp/4x/test.jpg',
               ["scale_in 3200 2400 True"]),
              ]),

            ('<img class="image-process-crisp2" src="/tmp/test.jpg" />',
             '<img class="image-process-crisp2" '
             'src="/tmp/derivs/crisp2/default/test.jpg" '
             'srcset="/tmp/derivs/crisp2/1x/test.jpg 1x, '
             '/tmp/derivs/crisp2/2x/test.jpg 2x, '
             '/tmp/derivs/crisp2/4x/test.jpg 4x"/>',

             # default must be first, because the process execute it first
             [('tmp/test.jpg', 'crisp2/default/test.jpg',
               ["scale_in 400 300 True"]),
              ('tmp/test.jpg', 'crisp2/1x/test.jpg',
               ["scale_in 800 600 True"]),
              ('tmp/test.jpg', 'crisp2/2x/test.jpg',
               ["scale_in 1600 1200 True"]),
              ('tmp/test.jpg', 'crisp2/4x/test.jpg',
               ["scale_in 3200 2400 True"])]),

            ('<img class="image-process-large-photo" src="/tmp/test.jpg" />',
             '<img class="image-process-large-photo" '
             'sizes="(min-width: 1200px) 800px, (min-width: 992px) 650px, '
             '(min-width: 768px) 718px, 100vw" src="/tmp/derivs/large-photo/'
             '800w/test.jpg" srcset="/tmp/derivs/large-photo/600w/test.jpg '
             '600w, /tmp/derivs/large-photo/800w/test.jpg 800w, '
             '/tmp/derivs/large-photo/1600w/test.jpg 1600w"/>',
             [('tmp/test.jpg', 'large-photo/600w/test.jpg',
               ["scale_in 600 450 True"]),
              ('tmp/test.jpg', 'large-photo/800w/test.jpg',
               ["scale_in 800 600 True"]),
              ('tmp/test.jpg', 'large-photo/1600w/test.jpg',
               ["scale_in 1600 1200 True"]),
              ]),
        ]

        for data in test_data:
            html = harvest_images_in_fragment(data[0], settings)

            expected_images = []
            expected_calls = []

            for t in data[2]:
                expected_source = os.path.join(settings['PATH'], t[0])
                expected_destination = os.path.join(
                    settings['OUTPUT_PATH'], 'tmp',
                    settings['IMAGE_PROCESS_DIR'], t[1]
                )

                expected_image = (expected_source, expected_destination, t[2])
                expected_images.append(expected_image)
                expected_calls.append(mock.call(expected_image, settings))

            self.maxDiff = None
            self.assertEqual(html, data[1])
            self.assertEqual(process_image.call_args_list, expected_calls)

            process_image.reset_mock()

    @mock.patch('image_process.process_image')
    def test_picture_generation(self, process_image):
        settings = get_settings(IMAGE_PROCESS=self.valid_transforms,
                                IMAGE_PROCESS_DIR='derivs')
        test_data = [
            ('<picture><source class="source-1" '
             'src="/images/pelican-closeup.jpg"></source><img '
             'class="image-process-pict" src="/images/pelican.jpg"/>'
             '</picture>',

             '<picture><source media="(min-width: 640px)" sizes="100vw" '
             'srcset="/images/derivs/pict/default/640w/pelican.jpg 640w, '
             '/images/derivs/pict/default/1024w/pelican.jpg 1024w, '
             '/images/derivs/pict/default/1600w/pelican.jpg 1600w">'
             '</source><source srcset="/images/derivs/pict/source-1/1x/'
             'pelican-closeup.jpg 1x, /images/derivs/pict/source-1/2x/'
             'pelican-closeup.jpg 2x"></source><img class="image-process-pict"'
             ' src="/images/derivs/pict/default/640w/pelican.jpg"/></picture>',

             [('images/pelican.jpg', 'pict/default/640w/pelican.jpg',
                 ["scale_in 640 480 True"]),
              ('images/pelican.jpg', 'pict/default/1024w/pelican.jpg',
                 ["scale_in 1024 683 True"]),
              ('images/pelican.jpg', 'pict/default/1600w/pelican.jpg',
                 ["scale_in 1600 1200 True"]),
              ('images/pelican-closeup.jpg',
                  'pict/source-1/1x/pelican-closeup.jpg',
                  ["crop 100 100 200 200"]),
              ('images/pelican-closeup.jpg',
                  'pict/source-1/2x/pelican-closeup.jpg',
                  ["crop 100 100 300 300"])
              ]),

            ('<div class="figure"><img alt="Pelican" class="image-process'
             '-pict2" src="/images/pelican.jpg" /> <p class="caption">'
             'A nice pelican</p> <div class="legend"> <img alt="Other view of '
             'pelican" class="image-process source-2" src="/images/'
             'pelican-closeup.jpg" /></div></div>',

             '<div class="figure"><picture><source media="(min-width: 640px)" '
             'sizes="100vw" srcset="/images/derivs/pict2/default/640w/pelican'
             '.jpg 640w, /images/derivs/pict2/default/1024w/pelican.jpg 1024w,'
             ' /images/derivs/pict2/default/1600w/pelican.jpg 1600w"></source>'
             '<source srcset="/images/derivs/pict2/source-2/1x/pelican-closeup'
             '.jpg 1x, /images/derivs/pict2/source-2/2x/pelican-closeup.jpg '
             '2x"></source><img alt="Pelican" class="image-process-pict2" '
             'src="/images/derivs/pict2/source-2/default/pelican-closeup.jpg"'
             '/></picture> <p class="caption">A nice pelican</p> <div '
             'class="legend"> </div></div>',

             [
                # default call first
                ('images/pelican-closeup.jpg',
                 'pict2/source-2/default/pelican-closeup.jpg',
                 ["scale_in 800 600 True"]),

                ('images/pelican.jpg', 'pict2/default/640w/pelican.jpg',
                 ["scale_in 640 480 True"]),

                # then images in order of processing
                ('images/pelican.jpg', 'pict2/default/1024w/pelican.jpg',
                 ["scale_in 1024 683 True"]),
                ('images/pelican.jpg', 'pict2/default/1600w/pelican.jpg',
                 ["scale_in 1600 1200 True"]),

                ('images/pelican-closeup.jpg',
                 'pict2/source-2/1x/pelican-closeup.jpg',
                 ["crop 100 100 200 200"]),
                ('images/pelican-closeup.jpg',
                 'pict2/source-2/2x/pelican-closeup.jpg',
                 ["crop 100 100 300 300"])
              ])
           ]

        for data in test_data:
            html = harvest_images_in_fragment(data[0], settings)

            expected_images = []
            expected_calls = []
            for t in data[2]:
                expected_source = os.path.join(settings['PATH'], t[0])
                expected_destination = os.path.join(
                    settings['OUTPUT_PATH'], 'images',
                    settings['IMAGE_PROCESS_DIR'], t[1]
                )

                expected_image = (expected_source, expected_destination, t[2])
                expected_images.append(expected_image)
                expected_calls.append(mock.call(expected_image, settings))

            self.maxDiff = None
            self.assertEqual(html, data[1])
            self.assertEqual(process_image.call_args_list, expected_calls)

            process_image.reset_mock()


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*- #
"""
Image Process
=============

This plugin process images according to their class attribute.
"""

from __future__ import unicode_literals

import copy
import collections
import functools
import os.path
import re
import six

from six.moves.urllib_parse import urlparse, urljoin, unquote
from six.moves.urllib_request import url2pathname, pathname2url
from PIL import Image, ImageFilter
from bs4 import BeautifulSoup
from pelican import signals

__version__ = "1.1.0"

IMAGE_PROCESS_REGEX = re.compile("image-process-[-a-zA-Z0-9_]+")

Path = collections.namedtuple(
    'Path', ['base_url', 'source', 'base_path', 'filename', 'process_dir']
)


def convert_box(image, t, l, r, b):
    """Convert box coordinates strings to integer.

    t, l, r, b (top, left, right, bottom) must be strings specifying
    either a number or a percentage.
    """
    bbox = image.getbbox()
    iw = bbox[2] - bbox[0]
    ih = bbox[3] - bbox[1]

    if t[-1] == '%':
        t = ih * float(t[:-1]) / 100.
    else:
        t = float(t)
    if l[-1] == '%':
        l = iw * float(l[:-1]) / 100.
    else:
        l = float(l)
    if r[-1] == '%':
        r = iw * float(r[:-1]) / 100.
    else:
        r = float(r)
    if b[-1] == '%':
        b = ih * float(b[:-1]) / 100.
    else:
        b = float(b)

    return (t, l, r, b)


def crop(i, t, l, r, b):
    """Crop image i to the box (l,t)-(r,b).

    t, l, r, b (top, left, right, bottom) must be strings specifying
    either a number or a percentage.
    """
    t, l, r, b = convert_box(i, t, l, r, b)
    return i.crop((int(t), int(l), int(r), int(b)))


def resize(i, w, h):
    """Resize the image to the dimension specified.

    w, h (width, height) must be strings specifying either a number
    or a percentage.
    """
    _, _, w, h = convert_box(i, '0', '0', w, h)

    if i.mode == 'P':
        i = i.convert('RGBA')
    elif i.mode == '1':
        i = i.convert('L')

    return i.resize((int(w), int(h)), Image.LANCZOS)


def scale(i, w, h, upscale, inside):
    """Resize the image to the dimension specified, keeping the aspect
    ratio.

    w, h (width, height) must be strings specifying either a number
    or a percentage, or "None" to ignore this constraint.

    If upscale is True, upscaling is allowed.

    If inside is True, the resulting image will not be larger than the
    dimensions specified, else it will not be smaller.
    """
    bbox = i.getbbox()
    iw = bbox[2] - bbox[0]
    ih = bbox[3] - bbox[1]

    if w == 'None':
        w = 1.
    elif w[-1] == '%':
        w = float(w[:-1]) / 100.
    else:
        w = float(w) / iw

    if h == 'None':
        h = 1.
    elif h[-1] == '%':
        h = float(h[:-1]) / 100.
    else:
        h = float(h) / ih

    if inside:
        scale = min(w, h)
    else:
        scale = max(w, h)

    if upscale in [0, '0', 'False', False]:
        scale = min(scale, 1.)

    if i.mode == 'P':
        i = i.convert('RGBA')
    elif i.mode == '1':
        i = i.convert('L')

    return i.resize((int(scale*iw), int(scale*ih)), Image.LANCZOS)


def rotate(i, degrees):
    if i.mode == 'P':
        i = i.convert('RGBA')
    elif i.mode == '1':
        i = i.convert('L')

    # rotate does not support the LANCZOS filter (Pillow 2.7.0).
    return i.rotate(int(degrees), Image.BICUBIC, True)


def apply_filter(i, f):
    if i.mode == 'P':
        i = i.convert('RGBA')
    elif i.mode == '1':
        i = i.convert('L')

    return i.filter(f)


basic_ops = {
    'crop': crop,
    'flip_horizontal': lambda i: i.transpose(Image.FLIP_LEFT_RIGHT),
    'flip_vertical': lambda i: i.transpose(Image.FLIP_TOP_BOTTOM),
    'grayscale': lambda i: i.convert('L'),
    'resize': resize,
    'rotate': rotate,
    'scale_in': functools.partial(scale, inside=True),
    'scale_out': functools.partial(scale, inside=False),

    'blur': functools.partial(apply_filter, f=ImageFilter.BLUR),
    'contour': functools.partial(apply_filter, f=ImageFilter.CONTOUR),
    'detail': functools.partial(apply_filter, f=ImageFilter.DETAIL),
    'edge_enhance': functools.partial(apply_filter,
                                      f=ImageFilter.EDGE_ENHANCE),
    'edge_enhance_more': functools.partial(apply_filter,
                                           f=ImageFilter.EDGE_ENHANCE_MORE),
    'emboss': functools.partial(apply_filter, f=ImageFilter.EMBOSS),
    'find_edges': functools.partial(apply_filter, f=ImageFilter.FIND_EDGES),
    'smooth': functools.partial(apply_filter, f=ImageFilter.SMOOTH),
    'smooth_more': functools.partial(apply_filter, f=ImageFilter.SMOOTH_MORE),
    'sharpen': functools.partial(apply_filter, f=ImageFilter.SHARPEN),
}


def harvest_images(path, context):
    # Set default value for 'IMAGE_PROCESS_DIR'.
    if 'IMAGE_PROCESS_DIR' not in context:
        context['IMAGE_PROCESS_DIR'] = 'derivatives'

    # Set default value for 'IMAGE_PROCESS_ENCODING'
    if 'IMAGE_PROCESS_ENCODING' not in context:
        context['IMAGE_PROCESS_ENCODING'] = 'utf-8'

    with open(path, 'r+', encoding=context['IMAGE_PROCESS_ENCODING']) as f:
        res = harvest_images_in_fragment(f, context)
        f.seek(0)
        f.truncate()
        f.write(res)


def harvest_images_in_fragment(fragment, settings):
    parser = settings.get("IMAGE_PROCESS_PARSER", "html.parser")
    soup = BeautifulSoup(fragment, parser)

    for img in soup.find_all('img', class_=IMAGE_PROCESS_REGEX):
        for c in img['class']:
            if c.startswith('image-process-'):
                derivative = c[14:]
                break
        else:
            continue

        try:
            d = settings['IMAGE_PROCESS'][derivative]
        except KeyError:
            raise RuntimeError('Derivative %s undefined.' % derivative)

        if isinstance(d, list):
            # Single source image specification.
            process_img_tag(img, settings, derivative)

        elif not isinstance(d, dict):
            raise RuntimeError('Derivative %s definition not handled'
                               '(must be list or dict)' % (derivative))

        elif 'type' not in d:
            raise RuntimeError('"type" is mandatory for %s.' % derivative)

        elif d['type'] == 'image':
            # Single source image specification.
            process_img_tag(img, settings, derivative)

        elif d['type'] == 'responsive-image':
            # srcset image specification.
            build_srcset(img, settings, derivative)

        elif d['type'] == 'picture':
            # Multiple source (picture) specification.
            group = img.find_parent()
            if group.name == 'div':
                convert_div_to_picture_tag(soup, img, group, settings,
                                           derivative)
            elif group.name == 'picture':
                process_picture(soup, img, group, settings, derivative)

    return str(soup)


def compute_paths(img, settings, derivative):
    process_dir = settings['IMAGE_PROCESS_DIR']
    img_src = urlparse(img['src'])
    img_src_path = url2pathname(img_src.path[1:])
    img_src_dirname, filename = os.path.split(img_src_path)
    derviate_path = os.path.join(process_dir, derivative)
    base_url = urljoin(img_src.geturl(), pathname2url(derviate_path))

    for f, contobj in settings['filenames'].items():
        if img_src_path.endswith(contobj.save_as):
            source = contobj.source_path
            base_path = os.path.join(contobj.settings['OUTPUT_PATH'],
                                     os.path.dirname(contobj.save_as),
                                     process_dir, derivative)
            break
    else:
        site_url = urlparse(settings["SITEURL"])
        site_url_path = url2pathname(site_url.path[1:])
        # if SITEURL is undefined, don't break!
        if site_url_path:
            src_path = img_src_path.partition(site_url_path)[2].lstrip("/")
        else:
            src_path = img_src_path.lstrip("/")
        source = os.path.join(settings['PATH'], src_path)
        base_path = os.path.join(settings['OUTPUT_PATH'],
                                 os.path.dirname(src_path),
                                 derviate_path)

    return Path(base_url, source, base_path, filename, process_dir)


def process_img_tag(img, settings, derivative):
    path = compute_paths(img, settings, derivative)
    process = settings['IMAGE_PROCESS'][derivative]

    img['src'] = os.path.join(path.base_url, path.filename)
    destination = os.path.join(path.base_path, path.filename)

    if not isinstance(process, list):
        process = process['ops']

    process_image((path.source, destination, process), settings)


def build_srcset(img, settings, derivative):
    path = compute_paths(img, settings, derivative)
    process = settings['IMAGE_PROCESS'][derivative]

    default = process['default']
    if isinstance(default, six.string_types):
        default_name = default
    elif isinstance(default, list):
        default_name = 'default'
        destination = os.path.join(path.base_path, default_name, path.filename)
        process_image((path.source, destination, default), settings)

    img['src'] = os.path.join(path.base_url, default_name, path.filename)

    if 'sizes' in process:
        img['sizes'] = process['sizes']

    srcset = []
    for src in process['srcset']:
        file_path = os.path.join(path.base_url, src[0], path.filename)
        srcset.append("%s %s" % (file_path, src[0]))
        destination = os.path.join(path.base_path, src[0], path.filename)
        process_image((path.source, destination, src[1]), settings)

    if len(srcset) > 0:
        img['srcset'] = ', '.join(srcset)


def convert_div_to_picture_tag(soup, img, group, settings, derivative):
    """
    Convert a div containing multiple images to a picture.
    """
    process_dir = settings['IMAGE_PROCESS_DIR']
    # Compile sources URL. Special source "default" uses the main
    # image URL. Other sources use the img with classes
    # [source['name'], 'image-process'].  We also remove the img from
    # the DOM.
    sources = copy.deepcopy(settings['IMAGE_PROCESS'][derivative]['sources'])
    for s in sources:
        if s['name'] == 'default':
            s['url'] = img['src']
        else:
            candidates = group.find_all('img', class_=s['name'])
            for candidate in candidates:
                if 'image-process' in candidate['class']:
                    s['url'] = candidate['src']
                    candidate.decompose()
                    break

        url_path, s['filename'] = os.path.split(s['url'])
        s['base_url'] = os.path.join(url_path, process_dir, derivative)
        s['base_path'] = os.path.join(settings['OUTPUT_PATH'],
                                      s['base_url'][1:])

    # If default is not None, change default img source to the image
    # derivative referenced.
    default = settings['IMAGE_PROCESS'][derivative]['default']
    if default is not None:
        default_source_name = default[0]

        default_source = None
        for s in sources:
            if s['name'] == default_source_name:
                default_source = s
                break

        if default_source is None:
            raise RuntimeError(
                'No source matching "%s", referenced in default setting.',
                (default_source_name,)
            )

        if isinstance(default[1], six.string_types):
            default_item_name = default[1]

        elif isinstance(default[1], list):
            default_item_name = 'default'

            source = os.path.join(settings['PATH'], default_source['url'][1:])
            destination = os.path.join(s['base_path'], default_source_name,
                                       default_item_name,
                                       default_source['filename'])
            process_image((source, destination, default[1]), settings)

        # Change img src to url of default processed image.
        img['src'] = os.path.join(s['base_url'], default_source_name,
                                  default_item_name,
                                  default_source['filename'])

    # Create picture tag.
    picture_tag = soup.new_tag('picture')
    for s in sources:
        # Create new <source>
        source_attrs = {k: s[k] for k in s if k in ['media', 'sizes']}
        source_tag = soup.new_tag('source', **source_attrs)

        srcset = []
        for src in s['srcset']:
            srcset.append("%s %s" % (os.path.join(s['base_url'], s['name'],
                                     src[0], s['filename']), src[0]))

            source = os.path.join(settings['PATH'], s['url'][1:])
            destination = os.path.join(s['base_path'], s['name'], src[0],
                                       s['filename'])
            process_image((source, destination, src[1]), settings)

        if len(srcset) > 0:
            source_tag['srcset'] = ', '.join(srcset)

        picture_tag.append(source_tag)

    # Wrap img with <picture>
    img.wrap(picture_tag)


def process_picture(soup, img, group, settings, derivative):
    """
    Convert a simplified picture to a full HTML picture:

    <picture>
    <source class="source-1" src="image1.jpg"></source>
    <source class="source-2" src="image2.jpg"></source>
    <img class="image-process-picture" src="image3.jpg"></img>
    </picture>

    to

    <picture>
    <source srcset="...image1.jpg..." media="..." sizes="..."></source>
    <source srcset="...image2.jpg..."></source>
    <source srcset="...image3.jpg..." media="..." sizes="..."></source>
    <img src=".../image3.jpg"></img>
    </picture>

    """

    process_dir = settings['IMAGE_PROCESS_DIR']
    process = settings['IMAGE_PROCESS'][derivative]
    # Compile sources URL. Special source "default" uses the main
    # image URL. Other sources use the <source> with classes
    # source['name'].  We also remove the <source>s from the DOM.
    sources = copy.deepcopy(process['sources'])
    for s in sources:
        if s['name'] == 'default':
            s['url'] = img['src']
            source_attrs = {k: s[k] for k in s if k in ['media', 'sizes']}
            s['element'] = soup.new_tag('source', **source_attrs)
        else:
            s['element'] = group.find('source', class_=s['name']).extract()
            s['url'] = s['element']['src']
            del s['element']['src']
            del s['element']['class']

        url_path, s['filename'] = os.path.split(s['url'])
        s['base_url'] = os.path.join(url_path, process_dir, derivative)
        s['base_path'] = os.path.join(settings['OUTPUT_PATH'],
                                      s['base_url'][1:])

    # If default is not None, change default img source to the image
    # derivative referenced.
    default = process['default']
    if default is not None:
        default_source_name = default[0]

        default_source = None
        for s in sources:
            if s['name'] == default_source_name:
                default_source = s
                break

        if default_source is None:
            raise RuntimeError(
                'No source matching "%s", referenced in default setting.',
                (default_source_name,)
            )

        if isinstance(default[1], six.string_types):
            default_item_name = default[1]

        elif isinstance(default[1], list):
            default_item_name = 'default'
            source = os.path.join(settings['PATH'], default_source['url'][1:])
            destination = os.path.join(s['base_path'], default_source_name,
                                       default_item_name,
                                       default_source['filename'])

            process_image((source, destination, default[1]), settings)

        # Change img src to url of default processed image.
        img['src'] = os.path.join(s['base_url'], default_source_name,
                                  default_item_name,
                                  default_source['filename'])

    # Generate srcsets and put back <source>s in <picture>.
    for s in sources:
        srcset = []
        for src in s['srcset']:
            srcset.append("%s %s" % (os.path.join(s['base_url'], s['name'],
                                     src[0], s['filename']), src[0]))

            source = os.path.join(settings['PATH'], s['url'][1:])
            destination = os.path.join(s['base_path'], s['name'], src[0],
                                       s['filename'])
            process_image((source, destination, src[1]), settings)

        if len(srcset) > 0:
            # Append source elements to the picture in the same order
            # as they are found in
            # settings['IMAGE_PROCESS'][derivative]['sources'].
            s['element']['srcset'] = ', '.join(srcset)
            img.insert_before(s['element'])


def process_image(image, settings):
    # Set default value for 'IMAGE_PROCESS_FORCE'.
    if 'IMAGE_PROCESS_FORCE' not in settings:
        settings['IMAGE_PROCESS_FORCE'] = False

    # remove URL encoding to get to physical filenames
    image = list(image)
    image[0] = unquote(image[0])
    image[1] = unquote(image[1])
    # image[2] is the transformation

    path, _ = os.path.split(image[1])
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == 17:
            # Already exists
            pass

    # If original image is older than existing derivative, skip
    # processing to save time, unless user explicitely forced
    # image generation.
    if (settings['IMAGE_PROCESS_FORCE'] or
        not os.path.exists(image[1]) or
            os.path.getmtime(image[0]) > os.path.getmtime(image[1])):

        i = Image.open(image[0])

        for step in image[2]:
            if hasattr(step, '__call__'):
                i = step(i)
            else:
                elems = step.split(' ')
                i = basic_ops[elems[0]](i, *(elems[1:]))

        i.save(image[1])


def register():
    signals.content_written.connect(harvest_images)




import os
import re
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

base_dir = os.path.dirname(os.path.abspath(__file__))


def get_version(filename="minchin/pelican/plugins/image_process.py"):
    with open(os.path.join(base_dir, filename), encoding="utf-8") as initfile:
        for line in initfile.readlines():
            m = re.match("__version__ *= *['\"](.*)['\"]", line)
            if m:
                return m.group(1)

setup(
    name="minchin.pelican.plugins.image_process",
    version=get_version(),
    description="Pelican plugin for automating image processing. Written in Python.",
    long_description="\n\n".join([open(os.path.join(base_dir, "README.rst")).read()]),
    author="W. Minchin",
    author_email="w_minchin@hotmail.com",
    url="https://github.com/MinchinWeb/minchin.pelican.plugins.image_process",
    packages=find_packages(),
    namespace_packages=['minchin',
                        'minchin.pelican',
                        'minchin.pelican.plugins',
                        ],
    include_package_data=True,
    install_requires=[
        'pillow',
        'beautifulsoup4',
        'six',
        'pelican',
        ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
)
